/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2016-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.service

import com.persagy.base.extensions.rootCause
import com.persagy.base.extensions.toJson
import com.persagy.database.SAbstractDao
import com.persagy.database.SFilter
import com.persagy.database.extensions.keyMap
import com.persagy.service.models.enums.SResponseType
import com.persagy.service.models.requests.*
import com.persagy.service.models.responses.*
import org.slf4j.LoggerFactory
import org.springframework.web.bind.annotation.RequestBody
import kotlin.collections.ArrayList

/**
 * 对象访问服务类基类
 *
 * @author 庞利祥 <sybotan@126.com>
 */
open class SObjectService<ENTITY: Any>
    : SBaseService<ENTITY> {
    /** 伴生对象 */
    companion object {
        /** 日志记录器 */
        private val logger = LoggerFactory.getLogger(SObjectService::class.java)
    }

    /**
     * 构造函数
     *
     * @param entityClazz     实体类
     */
    constructor(entityClazz: Class<ENTITY>): super(entityClazz)

    /**
     * 构造函数
     *
     * @param daoObject       数据访问对象
     */
    constructor(daoObject: SAbstractDao<ENTITY>): super(daoObject)

    /**
     * 创建对象
     *
     * @param request         请求对象
     * @return 创建结果应答信息
     */
    fun createList(@RequestBody request: SCreateRequest<ENTITY>): SCreateResponse<ENTITY> {
        return try {
            if (request.isMerge) {
                for (entity in request.content) {
                    onCreateBefore(entity)
                    replace(entity)
                }
            } else {
                for (entity in request.content) {
                    val onCreateBefore = onCreateBefore(entity)
                    if (onCreateBefore){
                        insert(entity)
                    }
                }
            }
            onCreateSuccess(request.content)
            SCreateResponse(request.content)
        } catch (e: Exception) {
            onCreateFailure(request.content)
            e.printStackTrace()
            SCreateResponse(SResponseType.failure, e.message!!)
        }
    }

    /**
     * 删除对象
     *
     * @param entityList       对象 ID 列表
     * @param withFilterList   公共过滤条件
     * @return 删除结果应答信息
     */
    fun deleteByKeysList(@RequestBody entityList: ArrayList<ENTITY>, withFilterList: ArrayList<SFilter>? = null): SBaseResponse {
         try {
            for (entity in entityList) {
                val filterList = ArrayList<SFilter>()
                entity.keyMap().forEach {
                    filterList.add(SFilter.eq(it.key, it.value))
                }

                if (filterList.isNullOrEmpty()){
                    return SBaseResponse(SResponseType.failure, "实体类缺少主键定义")
                }

                if (withFilterList != null) {
                    filterList.addAll(withFilterList)
                }

                onDeleteBefore(entity)

                if (!delete(filterList)) {
                    onDeleteFailure(entityList)
                    return SBaseResponse(SResponseType.failure, "删除失败！")
                }
            }
            onDeleteSuccess(entityList)
             return SBaseResponse(SResponseType.success)
        } catch (e: Exception) {
            e.printStackTrace()
             return SBaseResponse(SResponseType.failure, e.message!!)
        }
    }

    /**
     * 根据条件删除对象
     *
     * @param filterList  删除条件
     * @return 删除结果应答信息
     */
    fun batchDelete(@RequestBody filterList: ArrayList<SFilter>): SBaseResponse {
        return if (delete(filterList)) {
            SBaseResponse(SResponseType.success)
        } else {
            SBaseResponse(SResponseType.failure, "删除失败！")
        }
    }

    /**
     * 更新对象
     *
     * @param request         更新请求
     * @param withFilterList  公共过滤条件
     * @return 更新结果应答信息
     */
    fun updateList(@RequestBody request: SUpdateRequest<ENTITY>, withFilterList: ArrayList<SFilter>? = null): SBaseResponse {
        val entityList = request.content
        return try {
            for (entity in entityList!!) {
                val filterList = ArrayList<SFilter>()
                entity.keyMap().forEach {
                    filterList.add(SFilter.eq(it.key, it.value))
                }

                if (withFilterList != null) {
                    filterList.addAll(withFilterList)
                }

                onUpdateBefore(entity)
                if (!update(entity, request.nullItems, filterList)) {
                    onUpdateFailure(entityList)
                    return SBaseResponse(SResponseType.failure, "更新失败！")
                }
            }

            onUpdateSuccess(entityList)
            SBaseResponse(SResponseType.success)
        } catch (e: Exception){
            e.printStackTrace()
            SBaseResponse(SResponseType.failure, e.message!!)
        }
    }

    /**
     * 根据条件分页查询
     *
     * @param request         查询请求
     * @param withFilterList  公共过滤条件
     * @return 查询结果应答信息
     */
    fun pageQuery(@RequestBody request: SQueryRequest, withFilterList: ArrayList<SFilter>? = null): SQueryResponse<ENTITY> {
        return try {
            val queryBuilder = select(request.filters)
            if (null != withFilterList) {
                queryBuilder.filterList.addAll(withFilterList)
                logger.debug("filterList = ${queryBuilder.filterList.toJson()}")
            }

            /** 排序条件 */
            if (null != request.orders) {
                queryBuilder.order(request.orders!!)
            }

            /** 去重 */
            if (null != request.distinct && request.distinct!!) {
                queryBuilder.distinct()
            }

            /** 映射标志 */
            if (null != request.projection) {
                queryBuilder.projection(request.projection!!)
            }

            /** 分组统计 */
            if (null != request.group && !request.group!!.nameList.isNullOrEmpty()) {
                /** 分组属性名列表 */
                queryBuilder.group(*request.group!!.nameList.toArray(emptyArray<String>()))
                /** 是否统计个数 */
                queryBuilder.count(request.group!!.count ?: false)
                /** 求和 */
                if (!request.group!!.sumList.isNullOrEmpty()) {
                    queryBuilder.sum(*request.group!!.sumList.toArray(emptyArray<String>()))
                }
                /** 求平均值 */
                if (!request.group!!.avgList.isNullOrEmpty()) {
                    queryBuilder.avg(*request.group!!.avgList.toArray(emptyArray<String>()))
                }
                /** 取最大值 */
                if (!request.group!!.maxList.isNullOrEmpty()) {
                    queryBuilder.max(*request.group!!.maxList.toArray(emptyArray<String>()))
                }
                /** 取最小值 */
                if (!request.group!!.minList.isNullOrEmpty()) {
                    queryBuilder.min(*request.group!!.minList.toArray(emptyArray<String>()))
                }
            }

            onQueryBefore(queryBuilder)
            val pageInfo = queryBuilder.pageQuery(request.pageNumber,request.pageSize)
            /** 如果需要进行级联查询 */
            if (request.cascade != null && request.cascade!!.size > 0 &&
                    pageInfo.content != null && pageInfo.content!!.size > 0) {
                @Suppress("UNCHECKED_CAST")
                processCascadeQuery(pageInfo.content!! as ArrayList<Any>, request.cascade!!)
            }

            onQuerySuccess(pageInfo)
            SQueryResponse(pageInfo)
        } catch (e : Exception){
            e.printStackTrace()
            SQueryResponse(SResponseType.failure, e.rootCause()?.message ?: e.message ?: "未知异常！")
        }
    }

    /**
     * 根据条件统计
     *
     * @param request         查询条件请求
     * @param withFilterList  公共过滤条件
     * @return 查询结果应答信息
     */
    fun count(@RequestBody request: SCountRequest, withFilterList: ArrayList<SFilter>? = null): SCountResponse {
        return try {
            val queryBuilder = select(request.filters)

            if (null != withFilterList) {
                queryBuilder.filterList.addAll(withFilterList)
            }

            onCountBefore(queryBuilder)
            val count = queryBuilder.execCount()
            onCountSuccess(count)
            SCountResponse(count)
        } catch (e : Exception){
            e.printStackTrace()
            SCountResponse(0, SResponseType.failure, e.rootCause()?.message ?: e.message ?: "未知异常！")
        }
    }
}
