/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2016-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.service

import org.slf4j.LoggerFactory
import javax.servlet.http.Cookie
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.servlet.http.HttpSession

/**
 * 页面上下文
 *
 * @author 庞利祥 <sybotan@126.com>
 */
object SPageContext {
    /** 日志记录器 */
    private val logger = LoggerFactory.getLogger(SPageContext::class.java)
    /** 客户端请求对象 */
    private val request = ThreadLocal<HttpServletRequest>()
    /** 服务器应答对象 */
    private val response = ThreadLocal<HttpServletResponse>()

    var domain = "sybotan"

    /**
     * 获得客户端请求对象
     *
     * @return 返回客户端请求对象
     */
    fun getRequest(): HttpServletRequest {
        return request.get()
    }

    /**
     * 设置客户端请求对象
     *
     * @param req 客户端请求对象
     */
    fun setRequest(req: HttpServletRequest) {
        request.set(req)
        return
    }

    /**
     * 释放保存客户端请求对象的变量
     */
    fun removeRequest() {
        request.remove()
        return
    }

    /**
     * 获得服务器应答对象
     *
     * @return 服务器应答对象
     */
    fun getResponse(): HttpServletResponse {
        return response.get()
    }

    /**
     * 设置服务器应答对象
     *
     * @param resp 服务器应答对象
     */
    fun setResponse(resp: HttpServletResponse) {
        response.set(resp)
        return
    }

    /**
     * 释放保存服务器应答对象的变量
     */
    fun removeResponse() {
        response.remove()
        return
    }

    /**
     * 获得客户端会话对象
     *
     * @return 返回客户端会话对象
     */
    fun getSession(): HttpSession? {
        val req = request.get()
        return req?.session
    }

    /**
     * 获得保存在客户端会话中的指定属性
     *
     * @param name   属性名
     * @return 返回保存在客户端会话中的指定属性
     */
    fun getAttribute(name: String): Any? {
        val session = getSession()
        return session?.getAttribute(name)
    }

    /**
     * 设置保存在客户端会话中的指定属性
     *
     * @param name   属性名
     * @param object 值
     */
    fun setAttribute(name: String, `object`: Any) {
        val session = getSession()
        session?.setAttribute(name, `object`)
        return
    }

    /**
     * 移除保存在客户端会话中的指定属性
     *
     * @param name 属性名
     */
    fun removeAttribute(name: String) {
        val session = getSession()
        session?.removeAttribute(name)
        return
    }

    /**
     * 移除保存在客户端会话中的所有属性
     */
    fun removeAllAttribute() {
        val session = getSession() ?: return

        val nameList = session.attributeNames
        while (nameList.hasMoreElements()) {
            val name = nameList.nextElement()
            session.removeAttribute(name)
        }
        return
    }

    /**
     * 获得cookie
     *
     * @param name cookie名
     * @return cookie
     */
    fun getCookie(name: String): Cookie? {
        val request = getRequest()
        val cookieList = request.cookies

        if (null != cookieList) {
            for (cookie in cookieList) {
                if (name.compareTo(cookie.name) == 0) {
                    return cookie
                }
            }
        }

        return null
    }

    /**
     * 添加cookie
     *
     * @param name cookie名称
     * @param value cookie值
     */
    fun setCookie(name: String, value: String) {
        val response = getResponse()
        val cookie = Cookie(name, value)

        cookie.domain = domain
        cookie.path = "/"
        cookie.maxAge = 365 * 24 * 3600     // cookie最大生命周期1年

        response.addCookie(cookie)
        return
    }

    /**
     * 删除cookie
     *
     * @param name cookie名称
     */
    fun removeCookie(name: String) {
        val response = getResponse()
        val cookie = Cookie(name, null)

        cookie.domain = domain
        cookie.path = "/"
        cookie.maxAge = 0
        response.addCookie(cookie)

        return
    }

    /**
     * 获得请求头
     *
     * @param   name    请求头名
     * @return  请求头内容
     */
    fun getHeader(name: String): String? {
        val request = getRequest()
        return request.getHeader(name)
    }

    /**
     * 获得客户端的IP地址
     *
     * @return 客户端的IP地址
     */
    fun getRequestIp(): String {
        val request = getRequest()

        var ip: String? = request.getHeader("x-forwarded-for")
        if (null == ip || ip.isEmpty() || ip.equals("unknown", ignoreCase = true)) {
            ip = request.getHeader("Proxy-Client-IP")
        }

        if (null == ip || ip.isEmpty() || ip.equals("unknown", ignoreCase = true)) {
            ip = request.getHeader("WL-Proxy-Client-IP")
        }

        if (null == ip || ip.isEmpty() || ip.equals("unknown", ignoreCase = true)) {
            ip = request.getHeader("HTTP_CLIENT_IP")
        }

        if (null == ip || ip.isEmpty() || ip.equals("unknown", ignoreCase = true)) {
            ip = request.getHeader("HTTP_X_FORWARDED_FOR")
        }

        if (null == ip || ip.isEmpty() || ip.equals("unknown", ignoreCase = true)) {
            ip = request.remoteAddr
            if (ip!!.contains(":")) {
                ip = "127.0.0.1"
            }
        }

        return ip
    }
}