/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2016-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.service.kafka

import com.persagy.base.extensions.toJson
import org.springframework.kafka.core.KafkaTemplate
import org.springframework.kafka.support.SendResult
import org.springframework.util.concurrent.ListenableFuture
import org.springframework.util.concurrent.ListenableFutureCallback
import java.util.*

/**
 * Kafka 服务。用于连接 kafka, 收发 kafka 消息
 *
 * @author 庞利祥 <sybotan@126.com>
 */
open class SKafkaService {
    /** 延迟发送时间 */
    var delaySendTime = 3000L
    /** Kafka 访问模板 */
    @Suppress("UNCHECKED_CAST")
    lateinit var kafkaTemplate: KafkaTemplate<Int, String>
    /** 实现迟延相关数据 */
    private val delayMap = HashMap<String, SDelayData>()

    /**
     * 发送通知
     *
     * @param topic       发送标题
     * @param notice      通知
     * @param delaySend   是否启动延迟发送，默认延迟发送
     */
    fun sendNotice(topic: String, notice: SBaseNotice, delaySend: Boolean = true) {
        val delayData = delayMap[topic] ?: SDelayData()
        if (!delayMap.containsKey(topic)) {
            delayMap[topic] = delayData
        }

        /** 合并请求 */
        if (!mergeNotice(topic, notice)) {
            if (delayData.lastNotice != null) {
                sendTopic(topic, delayData.lastNotice!!.toJson())
            }
            delayData.lastNotice = notice
        }

        /** 是否延迟发送 */
        if (delaySend) {
            delaySend(topic)
        } else {
            flush(topic)
        }

        return
    }

    /**
     * 发送 topic 消息
     *
     * @param topic       发送标题
     * @param data        消息数据
     */
    fun sendTopic(topic: String, data: String) {
        val future: ListenableFuture<SendResult<Int, String>> = kafkaTemplate.send(topic, data).also {
            it.addCallback(object : ListenableFutureCallback<SendResult<Int, String>> {
                /**
                 * 发送失败回调
                 */
                override fun onFailure(ex: Throwable) {
                } // Fun onFailure
                /**
                 * 发送成功回调
                 */
                override fun onSuccess(result: SendResult<Int, String>) {
                } // Fun onSuccess()
            })
        }

        kafkaTemplate.flush()
    }

    /**
     * 合并通知
     *
     * @param lastNotice          缓存的通知
     * @param  notice              新的通知
     * @return 成功返回 true，否则返回 false
     */
    open fun mergeNotice(lastNotice: SBaseNotice, notice: SBaseNotice): Boolean {
        return false
    }

    /**
     * 延迟发送缓存的消息
     *
     * @param topic       发送标题
     */
    private fun delaySend(topic: String) {
        val delayData = delayMap[topic] ?: return

        /** 如果延迟任务不为空，则取消之前的延迟任务。保证延迟任务在最后一次发送后执行。 */
        if (delayData.delayTask != null) {
            delayData.delayTask!!.cancel()
        }

        /** 创建延迟任务 */
        delayData.delayTask = object : TimerTask() {
            override fun run() {
                /** 发送缓存的任务 */
                flush(topic)
            }
        }

        /** 启动延迟发送任务 */
        delayData.delayTimer.schedule(delayData.delayTask, delaySendTime)

        return
    }

    /**
     * 发送缓存的消息
     *
     * @param topic       发送标题
     */
    fun flush(topic: String) {
        val delayData = delayMap[topic] ?: return
        /** 如果没有缓存通知 */
        if (delayData.lastNotice == null) {
            return
        }

        sendTopic(topic, delayData.lastNotice!!.toJson())
        delayData.lastNotice = null

        return
    }

    /**
     * 合并通知
     *
     * @param topic       发送标题
     * @param notice      要合并的通知
     * @return 成功返回 true，否则返回 false
     */
    private fun mergeNotice(topic: String, notice: SBaseNotice): Boolean {
        val delayData = delayMap[topic] ?: return false

        /** 如果最后发送的通知为空 */
        if (delayData.lastNotice == null) {
            delayData.lastNotice = notice
            return true
        }

        return mergeNotice(delayData.lastNotice!!, notice)
    }

    /**
     * 延迟处理数据
     */
    inner class SDelayData {
        /** 最后发送的通知 */
        var lastNotice: SBaseNotice? = null
        /** 延迟发送通知定时器 */
        val delayTimer = Timer()
        /** 延迟发送通知定时任务 */
        var delayTask: TimerTask? = null
    }
}
